"""
Random Atom Factory
===================

"""

import random
import itertools as it

from .atom_factory import AtomFactory


class RandomAtomFactory(AtomFactory):
    """
    Creates a random collection of atoms.

    """

    def __init__(
        self,
        atoms,
        num_atoms,
        required_atoms=(),
        random_seed=None,
    ):
        """
        Initialize a :class:`.RandomAtomFactory`.

        Parameters
        ----------
        atoms : :class:`tuple` of :class:`.Atom`
            The atoms, which are randomly selected, with replacement,
            in each :meth:`~.RandomAtomFactory.get_atoms` call.

        num_atoms : :class:`int`
            The number of atoms which should be yielded in each
            :meth:`~.RandomAtomFactory.get_atoms` call.

        required_atoms : :class:`tuple` of :class:`.Atom`, optional
            Atoms, which must be yielded in each
            :meth:`~.RandomAtomFactory.get_atoms` call.

        random_seed : :class:`int`, optional
            The random seed to use.

        """

        self._atoms = atoms
        self._num_atoms = num_atoms
        self._required_atoms = required_atoms
        self._generator = random.Random(random_seed)

    def get_atoms(self):
        yield from it.islice(self._required_atoms, self._num_atoms)

        num_atoms = self._num_atoms - len(self._required_atoms)
        if num_atoms > 0 and self._atoms:
            yield from self._generator.choices(
                population=self._atoms,
                k=num_atoms,
            )



"""
Random Bond Factory
===================

"""

import random

from .utilities import ValenceTracker
from ..bond_factory import BondFactory
from ...bond import Bond


class RandomBondFactory(BondFactory):
    """
    Creates random collections of bonds.

    The bonds yielded will not necessarily create a single molecule.

    """

    def __init__(
        self,
        max_internal_bonds=10,
        required_bonds=(),
        random_seed=None,
    ):
        """
        Initialize a :class:`.RandomBondFactory` instance.

        Parameters
        ----------
        max_internal_bonds : :class:`int`, optional
            Once all atoms have been connected into a single molecule,
            the factory will create a random number of internal bonds
            between atoms in the molecule. This sets the maximum number
            of such bonds.

        required_bonds : :class:`tuple` of :class:`.Bond`
            Bonds, which must be yielded by each
            :meth:`~.BondFactory.get_bonds` call.

        random_seed : :class:`int`, optional
            The random seed to use.

        """

        self._max_internal_bonds = max_internal_bonds
        self._required_bonds = required_bonds
        self._generator = random.Random(random_seed)

    def get_bonds(self, atoms):
        yield from self._required_bonds

        valence_tracker = ValenceTracker(atoms, self._required_bonds)

        # To make sure a connected molecule is sampled, create two sets
        # of atoms. The first set of atoms holds atoms which have
        # been added to the molecule, while the second set of atom
        # holds atoms which have not yet been added to the molecule.

        # To randomly sample a connected molecule, first randomly
        # pick an atom from set 1 and then randomly pick an atom from
        # set 2. Yield a bond between these two atoms, and transfer
        # the atom from set 2 into set 1. Repeat this process until
        # set 2 is empty.
        while (
            valence_tracker.has_free_connected() and
            valence_tracker.get_num_disconnected() > 0
        ):
            atom1_id = self._generator.choice(
                seq=tuple(valence_tracker.get_free_connected()),
            )
            atom1_free_valence = (
                valence_tracker.get_free_valence(atom1_id)
            )

            atom2_id = self._generator.choice(
                seq=tuple(valence_tracker.get_disconnected()),
            )
            atom2_free_valence = (
                valence_tracker.get_free_valence(atom2_id)
            )
            orders = tuple(
                set(range(1, atom1_free_valence+1))
                &
                set(range(1, atom2_free_valence+1))
            )
            order = self._generator.choice(orders)
            bond = Bond(atom1_id, atom2_id, order)
            valence_tracker = valence_tracker.with_bond(bond)
            yield bond

        # Add bonds between connected atoms to form rings.
        num_internal_bonds = self._generator.randint(
            a=0,
            b=self._max_internal_bonds,
        )
        for i in range(num_internal_bonds):
            free_connected = tuple(
                valence_tracker.get_free_connected()
            )
            if len(free_connected) < 2:
                return

            atom1_id, atom2_id = self._generator.sample(
                population=free_connected,
                k=2,
            )
            atom1_free_valence = (
                valence_tracker.get_free_valence(atom1_id)
            )
            atom2_free_valence = (
                valence_tracker.get_free_valence(atom2_id)
            )
            orders = tuple(
                set(range(1, atom1_free_valence+1))
                &
                set(range(1, atom2_free_valence+1))
            )
            order = self._generator.choice(orders)
            bond = Bond(atom1_id, atom2_id, order)
            valence_tracker = valence_tracker.with_bond(bond)
            yield bond



"""
Atom Factory
============

"""


class AtomFactory:
    """
    Abstract base class for :class:`.Atom` factories.

    """

    # Keep empty __init__() to hide ugly default docstring.
    def __init__(self):
        """"""

    def get_atoms(self):
        """
        Yield the atoms of a molecule.

        Yields
        ------
        :class:`.Atom`
            An atom.

        """

        raise NotImplementedError()



"""
Bond
====

"""


class Bond:
    """
    A bond of a :class:`.Molecule`.

    """

    def __init__(self, atom1_id, atom2_id, order):
        """
        Initialize a :class:`.Bond` instance.

        Parameters
        ----------
        atom1_id : :class:`int`
            The id of the first atom of the bond.

        atom1_id : :class:`int`
            The id of the second atom of the bond.

        order : :class:`int`
            The order of the bond.

        """

        self._atom1_id = atom1_id
        self._atom2_id = atom2_id
        self._order = order

    def get_atom1_id(self):
        """
        Return the id of the first atom of the bond.

        Returns
        -------
        :class:`int`
            The id of the first atom of the bond.

        """

        return self._atom1_id

    def get_atom2_id(self):
        """
        Return the id of the second atom of the bond.

        Returns
        -------
        :class:`int`
            The id of the second atom of the bond.

        """

        return self._atom2_id

    def get_atom_ids(self):
        """
        Yield the ids of the atoms of the bond.

        Yields
        ------
        :class:`int`
            An atom id.

        """

        yield self._atom1_id
        yield self._atom2_id

    def get_order(self):
        """
        Return the bond order.

        Returns
        -------
        :class:`int`
            The bond order.

        """

        return self._order

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'{self._atom1_id}, {self._atom2_id}, '
            f'{self._order}'
            ')'
        )

    def __str__(self):
        return repr(self)



"""
Bond Factory
============

"""


class BondFactory:
    """
    Abstract base class for :class:`.Bond` factories.

    """

    # Keep empty __init__() to hide ugly default docstring.
    def __init__(self):
        """"""

    def get_bonds(self, atoms):
        """
        Yield the bonds of a molecule.

        Parameters
        ----------
        atoms : :class:`tuple` of :class:`.Atom`
            The atoms of the molecule.

        Yields
        ------
        :class:`.Bond`
            A bond.

        """

        raise NotImplementedError()



class ValenceTracker:
    def __init__(self, atoms, bonds=()):
        self._atoms = atoms
        self._free_valences = {
            atom_id: atom.get_max_valence()
            for atom_id, atom in enumerate(atoms)
        }
        self._connected = (
            frozenset() if bonds else frozenset({0})
        )
        self._disconnected = (
            frozenset(range(len(atoms)))
            if bonds
            else frozenset(range(1, len(atoms)))
        )
        for bond in bonds:
            self._with_bond(bond)

    def clone(self):
        clone = self.__class__.__new__(self.__class__)
        clone._atoms = self._atoms
        clone._free_valences = dict(self._free_valences)
        clone._connected = self._connected
        clone._disconnected = self._disconnected
        return clone

    def with_bond(self, bond):
        return self.clone()._with_bond(bond)

    def _with_bond(self, bond):
        connected = set()
        for atom_id in bond.get_atom_ids():
            connected.add(atom_id)
            self._free_valences[atom_id] = max(
                self._free_valences[atom_id] - bond.get_order(),
                0,
            )

        self._connected |= connected
        self._disconnected -= connected
        return self

    def get_ids_with_free_valence(self):
        yield from (
            atom_id
            for atom_id, free_valence in self._free_valences.items()
            if free_valence > 0
        )

    def get_free_valence(self, atom_id):
        return self._free_valences[atom_id]

    def get_free_connected(self):
        yield from (
            atom_id for atom_id in self._connected
            if self._free_valences[atom_id] > 0
        )

    def has_free_connected(self):
        return any(map(self._free_valences.get, self._connected))

    def get_num_disconnected(self):
        return len(self._disconnected)

    def get_disconnected(self):
        yield from self._disconnected



from setuptools import find_packages, setup
import re
from os.path import join


def get_version():
    with open(join('src', 'vabene', '__init__.py'), 'r') as f:
        content = f.read()
    p = re.compile(r'^__version__ = [\'"]([^\'\"]*)[\'"]', re.M)
    return p.search(content).group(1)


setup(
    name='vabene',
    author='Lukas Turcani',
    author_email='lukasturcani93@gmail.com',
    url='https://github.com/lukasturcani/vabene',
    version=get_version(),
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    install_requires=(),
    python_requires='>=3.3',
)



"""
Atom
====

"""


class Atom:
    """
    An atom of a molecule.

    """

    def __init__(self, atomic_number, charge, max_valence):
        """
        Initialize an :class:`.Atom` instance.

        Parameters
        ----------
        atomic_number : :class:`int`
            The atomic number of the atom.

        charge : :class:`int`
            The formal charge of the atom.

        max_valence : :class:`int`
            The maximum valence the atom is allowed to have.

        """

        self._atomic_number = atomic_number
        self._charge = charge
        self._max_valence = max_valence

    def get_atomic_number(self):
        """
        Return the atomic number of the atom.

        Returns
        -------
        :class:`int`
            The atomic number of the atom.

        """

        return self._atomic_number

    def get_charge(self):
        """
        Return the formal charge of the atom.

        Returns
        -------
        :class:`int`
            The formal charge of the atom.

        """

        return self._charge

    def get_max_valence(self):
        """
        Return the maximum valence the atom is allowed to have.

        Returns
        -------
        :class:`int`
            The maximum valence the atom is allowed to have.

        """

        return self._max_valence

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'{self._atomic_number}, {self._charge}, '
            f'{self._max_valence}'
            ')'
        )

    def __str__(self):
        return repr(self)



"""
Molecule
========

"""

from .utilities import dedupe


class Molecule:
    """
    A molecular graph.

    """

    def __init__(self, atoms, bonds):
        """
        Initialize a :class:`.Molecule` instance.

        Parameters
        ----------
        atoms : :class:`tuple` of :class:`.Atom`
            The atoms of the molecule.

        bonds : :class:`iterable` of :class:`.Bond`
            The bonds of the molecule.

        """

        self._atoms = atoms

        def get_atom_ids(bond):
            return frozenset(bond.get_atom_ids())

        self._bonds = tuple(dedupe(bonds, get_key=get_atom_ids))

    def get_atoms(self, atom_ids=None):
        """
        Yield the atoms of the molecule.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`
            The ids of the atoms which should be yielded. Can
            be an :class:`int`, if only one atom should
            be yielded.

        Yields
        ------
        :class:`.Atom`
            An atom of the molecule.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids, )

        for atom_id in atom_ids:
            yield self._atoms[atom_id]

    def get_num_atoms(self):
        """
        Return the number of atoms in the molecule.

        Returns
        -------
        :class:`int`
            The number of atoms in the molecule.

        """

        return len(self._atoms)

    def get_bonds(self, bond_ids=None):
        """
        Yield the bonds of the molecule.

        Parameters
        ----------
        bond_ids : :class:`iterable` of :class:`int`
            The ids of bonds which should be yielded. Can be an
            :class:`int`, if only one bond should be yielded.

        Yields
        ------
        :class:`.Bond`
            A bond of the molecule.

        """

        if bond_ids is None:
            bond_ids = range(len(self._bonds))
        elif isinstance(bond_ids, int):
            bond_ids = (bond_ids, )

        for bond_id in bond_ids:
            yield self._bonds[bond_id]

    def get_num_bonds(self):
        """
        Return the number of bonds in the molecule.

        Returns
        -------
        :class:`int`
            The number of bonds in the molecule.

        """

        return len(self._bonds)
